/**************************************************************************************************\
*                                        HEADER INCLUDES                                           *
\**************************************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include "emudefs.h"

/**************************************************************************************************\
*                                      MACRO DEFINITIONS                                           *
\**************************************************************************************************/
#define OBJFILE_HEADER_LENGTH 8
#define BYTE_SIZE 1
#define SHORT_SIZE 2
#define INT_SIZE 4

/**************************************************************************************************\
*                                         DATATYPE TYPEDEFS                                        *
\**************************************************************************************************/
typedef unsigned int uint; // Renamed for convenience

/**************************************************************************************************\
*                                     STRUCTURE DEFINITIONS                                        *
\**************************************************************************************************/

// Composite of segment locations
typedef struct
{
    uint ds_start_off; // Start of data segment in assembler file
    uint ds_end_off;   // End of data segment in assembler file
    uint cs_start_off; // Start of code segement in assembler file
} seglocator;

// Symbol table entry
typedef struct
{
    char* symbol;  // Name of symbol
    uint location; // Start location of symbol in object file
} symtable_entry;

// Symbol table
typedef struct
{
    symtable_entry* entries; // Array of symbol table entries
    uint numentries;         // Number of entries stored in table
} symtable;

// An array of strings (tokens)
typedef struct
{
    char** tokens; // Array of pointers to strings
    int numtokens; // Number of strings pointed to
} token_array;

/**************************************************************************************************\
*                                       FUNCTION PROTOTYPES                                        *
\**************************************************************************************************/
void        add_entry_to_symbol_table(char* symbol, uint location);
void        chop(char* line);
void        chomp(char* line);
char*       copy_string(char* str);
int         get_mem_address(char* token);
int         get_mem_address(char* token, address_mode_t* mode, register_t* reg);
register_t  get_register(char* reg);
int         get_datatype_size(char* type);
uint        get_symbol_location(char *label);
opcode_t    get_opcode(char *opcode);
void        get_segment_locations();
bool_t      is_constant(char* token, int* const_val);
bool_t      is_instruction(char* inst, opcode_t* opcode_save);
bool_t      is_mem_address(char* token, int* mem_address_save);
bool_t      is_register(char* token, register_t* reg_save);
void        parse_code();
void        parse_code_segment_pass1();
void        parse_code_segment_pass2();
void        parse_data_segment();
token_array tokenize_string(char* line, char* delims);
int         verify_instruction(char* inst);
void        write_objfile_header();

/**************************************************************************************************\
*                                        GLOBAL VARIABLES                                          *
\**************************************************************************************************/
FILE* asm_file;           // Handle to input assembler file
FILE* obj_file;           // Handle to output object file
symtable symbol_table;    // Symbol table
seglocator locator;       // Segment locator used to locate data & code segments in assembler file
int location;             // Location of file pointer in output file
int num_instructions = 0; // Number of instructions parsed
char curr_line[255];      // String representing the current line of assembler file being parsed

/**************************************************************************************************\
*                                          MAIN PROGRAM                                            *
\**************************************************************************************************/
int main(int argc, char *argv[])
{
    if (argc != 3) 
    {
        fprintf(stderr, "Usage: %s filename.asm filename.obj\n", argv[0]);
        return 1;
    }

    // Open Input and Output Files
    if ((asm_file = fopen(argv[1], "r")) == NULL) 
    {
        fprintf(stderr, "Unable to open %s\n", argv[1]);
        return 1;
    }

    if ((obj_file = fopen(argv[2], "wb")) == NULL) 
    {
        fprintf(stderr, "Unable to open %s\n",argv[2]);
        return 1;
    }

    // Clear memory for seglocator & symbol table
    memset(&symbol_table, 0, sizeof(symtable));
    memset(&locator, 0, sizeof(locator));

    // Parse code
    parse_code();

    // Close Input and Output Files
    fclose(asm_file);
    fclose(obj_file);

    return EXIT_SUCCESS;
}

/**************************************************************************************************\
*   FUNCTION NAME:  chomp
****************************************************************************************************
*
*   Description:    Removes one or more trailing new line characters from a string
*
*   Arguments:      1) Pointer to a C-styled string
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
void chomp(char* line)
{
    int i;
    for (i = strlen(line) - 1; i >=0 ; i-- )
    {
        if (line[i] == '\n')
        {
            line[i] = '\0';
        }
        else
        {
            break;
        }
    }
}

/**************************************************************************************************\
*   FUNCTION NAME:  get_segment_locations
****************************************************************************************************
*
*   Description:    Determines the byte offset within the assembler file of the data & code segments
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
void get_segment_locations()
{
    // Get line from file
    fgets(curr_line, 255, asm_file);

    while (!feof(asm_file))
    {
        chomp(curr_line);

        if (strcmp(curr_line, ".data") == 0)
        {
            locator.ds_start_off = ftell(asm_file);
        }
        else if (strcmp(curr_line, ".data_end") == 0)
        {
            locator.ds_end_off = ftell(asm_file);
        }
        else if (strcmp(curr_line, ".code") == 0)
        {
            locator.cs_start_off  = ftell(asm_file);
            break;
        }

        // Get next line
        fgets(curr_line, 255, asm_file);
    }
}

/**************************************************************************************************\
*   FUNCTION NAME:  copy_string
****************************************************************************************************
*
*   Description:    Makes a copy of a string
*
*   Arguments:      1) A pointer to a C-styled string
*
*   Return Value:   A pointer to a copy of the input string
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
char* copy_string(char* str)
{
    char* copy = (char*) malloc(strlen(str) + 1);
    strcpy(copy, str);
    return copy;
}

/**************************************************************************************************\
*   FUNCTION NAME:  tokenize_string
****************************************************************************************************
*
*   Description:    Parses a line for a set of tokens
*
*   Arguments:      1) Pointer to a C-styled string
*                   2) Pointer to a C-styled string of tokens
*
*   Return Value:   **
*
*   Special Notes:  None
*
*   Exceptions:     None 
*
\**************************************************************************************************/
token_array tokenize_string(char* line, char* delims)
{
    token_array tok_arr; // Token array to store tokens
    tok_arr.numtokens = 0;

    // Make copy of line
    char* linecpy = copy_string(line);
    char* token   = strtok(linecpy, delims);

    int i = 0, size = 0;
    for (; token != NULL ; i++)
    {
        if (tok_arr.numtokens == 0)
        {
            tok_arr.tokens = (char**) malloc(sizeof(char*));
            size = 1;
        }
        else
        {
            if (tok_arr.numtokens + 1 > size)
            {
                // Double size of array
                tok_arr.tokens = (char**) realloc(tok_arr.tokens, sizeof(char*) * size * 2);
                size *= 2;
            }
        }

        tok_arr.tokens[tok_arr.numtokens] = (char*) malloc(strlen(token) + 1);
        strcpy(tok_arr.tokens[tok_arr.numtokens], token);
        tok_arr.numtokens++;
        token = strtok(NULL, delims);
    }

    tok_arr.tokens = (char**) realloc(tok_arr.tokens, sizeof(char*) * tok_arr.numtokens);
    free(linecpy);
    return tok_arr;
}

/**************************************************************************************************\
*   FUNCTION NAME:  add_entry_to_symbol_table
****************************************************************************************************
*
*   Description:    Adds an entry to the symbol table
*
*   Arguments:      1) Pointer to C-styled string containing the symbol name
*                   2) The offset location of the symbol
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None 
*
\**************************************************************************************************/
void add_entry_to_symbol_table(char* symbol, uint location)
{
    symtable_entry entry;
    entry.location = location;
    entry.symbol   = copy_string(symbol);

    // Check if symbols are in the table
    if (symbol_table.numentries != 0) // One or more symbols in table
    {
        int found_in_table = 0;

        // Search table for pre-existing entry
        int i;
        for (i = 0; i < symbol_table.numentries; i++ )
        {
            if (strcmp(symbol_table.entries[i].symbol, symbol) == 0) // Symbol already exists
            {
                // Return old entry space to system
                free(symbol_table.entries[i].symbol);

                symbol_table.entries[i] = entry;
                found_in_table = 1;
                printf("Warning: Symbol %s has been redefined\n", symbol);
                break;
            }
        }

        if (found_in_table == 0) // Symbol not found in table, add it
        {
            symbol_table.entries = (symtable_entry*) realloc(symbol_table.entries,
                                   (symbol_table.numentries + 1) * sizeof(symtable_entry));
            symbol_table.entries[symbol_table.numentries++] = entry;
        }
    }
    else // No symbols in table
    {
        symbol_table.entries    = (symtable_entry*) malloc(sizeof(symtable_entry));
        symbol_table.entries[0] = entry;
        symbol_table.numentries++;
    }
}

/**************************************************************************************************\
*   FUNCTION NAME:  get_datatype_size
****************************************************************************************************
*
*   Description:    
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  Advances the file pointer of the assembler output file
*
*   Exceptions:     None 
*
\**************************************************************************************************/
int get_datatype_size(char* type)
{
    if (strcmp(type, ".byte") == 0)
    {
        return BYTE_SIZE;
    }
    else if (strcmp(type, ".short") == 0)
    {
        return SHORT_SIZE;
    }
    else if (strcmp(type, ".int") == 0)
    {
        return INT_SIZE;
    }
    else
    {
        return 0; // Inappropriate type specifier
    }
}

/**************************************************************************************************\
*   FUNCTION NAME:  parse_data_segment
****************************************************************************************************
*
*   Description:    Parses data segment of assembler input file
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  Advancese the file pointer of the assembler output file
*
*   Exceptions:     None 
*
\**************************************************************************************************/
void parse_data_segment()
{
    // Move pointer to start of data segment
    fseek(asm_file, locator.ds_start_off, SEEK_SET);

    // Move output pointer past header
    fseek(obj_file, OBJFILE_HEADER_LENGTH, SEEK_SET);

    while (ftell(asm_file) != locator.ds_end_off) // We have not reached end of data segement yet
    {
        // Get line from file
        fgets(curr_line, 255, asm_file);
        chomp(curr_line);

        // Break line into tokens
        token_array tok_arr = tokenize_string(curr_line, " ");

        if (tok_arr.numtokens >= 2)
        {
            int incr;
            if ( (incr = get_datatype_size(tok_arr.tokens[1])) == 0)
            {
                // Parse error
                return;
            }

            switch(tok_arr.numtokens)
            {
                case 2: // type varname
                {
                    add_entry_to_symbol_table(tok_arr.tokens[0], location);
                    break;
                }
                case 3: // type varname value
                {
                    add_entry_to_symbol_table(tok_arr.tokens[0], location);
                    break;
                }
            }

            // Increment location
            location += incr;
        }
        else
        {
            // Some kind of parse error
        }

        // Free pointer array
        //free_ptr_array(dl.tokens, dl.numtokens);
    }
}

/**************************************************************************************************\
*   FUNCTION NAME:  chop
****************************************************************************************************
*
*   Description:    Removes the last character of a string
*
*   Arguments:      1) A pointer to a C-styled string
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
void chop(char* line)
{
    line[strlen(line) - 1] = '\0';
}

/**************************************************************************************************\
*   FUNCTION NAME:  parse_code_segment_pass1
****************************************************************************************************
*
*   Description:    Performs the first pass of parsing the code segment. It records the symbols
*                   encountered during parsing.
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
void parse_code_segment_pass1()
{
    // Set file pointer to start of code segment
    fseek(asm_file, locator.cs_start_off, SEEK_SET);

    while (!feof(asm_file))
    {
         // Get line from file
        fgets(curr_line, 255, asm_file);
        chomp(curr_line);

        if (curr_line[strlen(curr_line) - 1] == ':') // Is a label
        {
            chop(curr_line);
            add_entry_to_symbol_table(curr_line, location);
        }
        else // Is an instruction
        {
            // Verify instruction and get it's length
            int instruction_len;
            //if ( (instruction_len = get_symbol_location(curr_line) > 0)
            
            location += MAX_INSTR_LENGTH;
        }
    }
}

/**************************************************************************************************\
*   FUNCTION NAME:  is_register
****************************************************************************************************
*
*   Description:    Determines whether or not an input string represents a register
*
*   Arguments:      1) Pointer to a C-styled string
*                   2) Pointer to a register_t that the result can be stored in.
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
bool_t is_register(char* token, register_t* reg_save)
{
    register_t reg = get_register(token);

    // Save result (if desired)
    if (reg_save != NULL)
    {
        *reg_save = reg;
    }

    if (reg != INVALID_REGISTER) // Is a register
    {
        return TRUE;
    }
    else // Is not a register
    {
        return FALSE;
    }
}

/**************************************************************************************************\
*   FUNCTION NAME:  is_mem_address
****************************************************************************************************
*
*   Description:    Determines whether or not an input string represents a memory address
*
*   Arguments:      1) A pointer to a C-styled string
*                   2) A pointer to an integer where the result be stored
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
bool_t is_mem_address(char* token, int* mem_address_save)
{
    int mem_address = get_mem_address(token);

    if (mem_address >= 0)
    {
        if (mem_address_save != NULL)
        {
            *mem_address_save = mem_address;
        }

        return TRUE;
    }
    else
    {
        return FALSE;
    }
}

/**************************************************************************************************\
*   FUNCTION NAME:  get_mem_address
****************************************************************************************************
*
*   Description:    
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
int get_mem_address(char* token)
{
    return get_mem_address(token, NULL, NULL);
}

/**************************************************************************************************\
*   FUNCTION NAME:  get_mem_address
****************************************************************************************************
*
*   Description:    
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
int get_mem_address(char* token, address_mode_t* mode, register_t* reg)
{
    int address;

    token_array tok_arr = tokenize_string(token, "[]+");
    switch(tok_arr.numtokens)
    {
        case 1:
        {
            address_mode_t mode_temp;

            if (is_register(tok_arr.tokens[0], reg) == TRUE) // Is a register (e.g. [R1])
            {
                mode_temp = INDIRECT;
            }
            else // Is a location (e.g. [1000])
            {
                // Clear errno before conversion
                errno = 0;
                address = atol(tok_arr.tokens[0]);

                mode_temp = DIRECT;
            }

            if (errno == 0)
            {
                if (mode != NULL)
                {
                    *mode = mode_temp;
                }

                return address;
            }
            else
            {
                return INVALID_MEMORY_ADDRESS;
            }
        }
        case 2:
        {
            // Indirect addressing
            if (mode != NULL)
            {
                *mode = INDEXED;
            }

            if (is_register(tok_arr.tokens[0], reg) == TRUE)
            {
                address = atol(tok_arr.tokens[1]);
                return address;

                // Check validity
            }
            else
            {
                if (is_register(tok_arr.tokens[1], reg) == TRUE)
                {
                    address = atol(tok_arr.tokens[0]);
                    return address;
                }
                else
                {
                    // Parse error
                }
            }
        }
    }

    return -1;
}

/**************************************************************************************************\
*   FUNCTION NAME:  get_address_mode
****************************************************************************************************
*
*   Description:    
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
address_mode_t get_address_mode(char* token)
{
    address_mode_t mode;

    if (is_constant(token, NULL) == TRUE)
    {
        return IMMEDIATE;
    }
    else if (is_mem_address(token, &mode) == TRUE)
    {
        return mode;
    }
    else if (is_register(token, NULL) == TRUE)
    {
        return REGISTER;
    }
}

/**************************************************************************************************\
*   FUNCTION NAME:  is_constant
****************************************************************************************************
*
*   Description:    
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
bool_t is_constant(char* token, int* const_val)
{
    // Clear errno before conversion
    errno = 0;
    int val = atol(token);

    bool_t rv = (errno == 0) : TRUE ? FALSE;

    if (const_val != NULL)
    {
        if (rv == TRUE)
        {
            *const_val = val;
        }
    }

    return rv;
}

/**************************************************************************************************\
*   FUNCTION NAME:  verify_instruction
****************************************************************************************************
*
*   Description:    
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
int verify_instruction(char* inst)
{
    token_array tok_arr = tokenize_string(inst, " ,");

    if (tok_arr.numtokens > 0)
    {
        opcode_t opcode = get_opcode(tok_arr.tokens);

        if (opcode == INVALID_OPCODE)
        {
            return -1;
        }

        switch(tok_arr.numtokens)
        {
/*            case 1: // Zero operand instruction
            {
                if (is_n_operand_instruction(opcode, 0))
                {
                    return 1;
                }
                else
                {
                    return -1;
                }
            }
            case 2: // One operand instruction
            {
                if (is_n_operand_instruction(opcode, 1))
                {
                    return get_operand_size(tok_arr.numtokens[1]);
                }
                else
                {
                    return -1;
                }
            }
            case 3:
            {
                if (is_n_operand_instruction(opcode, 2))
                {
                    return get_operand_size(tok_arr.numtokens[1]) + 
                           get_operand_size(tok_arr.numtokens[2]);
                }
                else
                {
                    return -1;
                }
            }
        }*/
    }
    else
    {
        /*
        // Check for blank line
        chomp(inst);

        if (strlen(inst) == 0) // Is blank line
        {
            return 0;
        }
        else
        {
            //
        }*/
    }
}

/**************************************************************************************************\
*   FUNCTION NAME:  is_instruction
****************************************************************************************************
*
*   Description:    
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
bool_t is_instruction(char* inst, opcode_t* opcode_save)
{
    opcode_t opcode = get_opcode(inst);

    // Save opcode
    if (opcode_save != NULL)
    {
        *opcode_save = opcode;
    }

    if (opcode != INVALID_OPCODE) // Is an instruction
    {
        return TRUE;
    }
    else // Is not an instruction
    {
        return FALSE;
    }
}

/**************************************************************************************************\
*   FUNCTION NAME:  parse_code_segment_pass2
****************************************************************************************************
*
*   Description:    Performs the second pass of parsing the code segment.
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
void parse_code_segment_pass2()
{
    // Set file pointer to start of code segment
    fseek(asm_file, locator.cs_start_off, SEEK_SET);
    opcode_t opcode;
    instruction instr;
    memset(instr, 0, sizeof(instruction));

     // Get line from file
    fgets(curr_line, 255, asm_file);

    while (!feof(asm_file))
    {
        chomp(curr_line);
        token_array tok_arr = tokenize_string(curr_line, " ,;");

        if (tok_arr.numtokens > 0)
        {
            opcode = get_opcode(tok_arr.tokens[0]);
            memset(instr, 0, sizeof(instruction));

            switch(opcode)
            {
                case MOV:
                case ADD:
                case SUB:
                case AND:
                case XOR:
                case OR:
                {
                    instr[0] = opcode;
                    instr[1] = get_register(tok_arr.tokens[1]);
                    instr[2] = get_register(tok_arr.tokens[2]);
                    break;
                }
                case MOVI:
                case ADDI:
                case SUBI:
                {
                    instr[0] = opcode;
                    instr[1] = get_register(tok_arr.tokens[1]);
                    short value = (short) atol(tok_arr.tokens[2]);
                    memcpy(&instr[2],&value, sizeof(short));
                    break;
                }
                case INC:
                case DEC:
                case NOT:
                {
                    instr[0] = opcode;
                    instr[1] = get_register(tok_arr.tokens[1]);
                    break;
                }
                case JMP:
                {
                    instr[0] = opcode;
                    short value = get_symbol_location(tok_arr.tokens[1]);
                    memcpy(&instr[1], &value, sizeof(short)); 
                    break;
                }
                case HLT:
                {
                    instr[0] = opcode;
                    break;
                }
            }

            fwrite(instr, MAX_INSTR_LENGTH, 1, obj_file);
            num_instructions++;
            fgets(curr_line, 255, asm_file);
        }
        else
        {
            // Parse error
        }
    }
}

/**************************************************************************************************\
*   FUNCTION NAME:  get_register
****************************************************************************************************
*
*   Description:    
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
register_t get_register(char* reg)
{
    char* register_strs[] = {"R0", "R1", "R2", "R3", "R4", "R5", "R6", "R7"};

    int i;
    for (i = 0; i < NUM_REGISTERS; i++ )
    {
        if (strcmp(register_strs[i], reg) == 0)
        {
            return (register_t) i;
        }
    }

    return INVALID_REGISTER;
}

/**************************************************************************************************\
*   FUNCTION NAME:  get_opcode
****************************************************************************************************
*
*   Description:    Looks up the opcode number by checking it's string representation
*
*   Arguments:      1) A pointer to a C-styled string
*
*   Return Value:   The opcode number OR INVALID_OPCODE
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
opcode_t get_opcode(char *opcode)
{
    char* opcode_strs[]  = {"ADD", "ADDI", "AND", "DEC", "DIV", "HLT", "INC", "JMP", "MOV", "MOVI",
                            "MUL", "NOT", "OR", "SUB", "SUBI", "XOR"};
    int i;
    for (i=0; i < NUM_OPCODES; i++ )
    {
        if (strcmp(opcode_strs[i], opcode) == 0)
        {
            return (opcode_t)(i + 1);
        }
    }

    return INVALID_OPCODE;
}

/**************************************************************************************************\
*   FUNCTION NAME:  get_symbol_location
****************************************************************************************************
*
*   Description:    Looks up an absolute location in the object file by symbol name
*
*   Arguments:      1) Pointer to a C-styled string representing the symbol name
*
*   Return Value:   Absolute location of symbol in object file
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
uint get_symbol_location(char* label)
{
    int i;
    for(i = 0; i < symbol_table.numentries; i++)
    {
        if (strcmp(symbol_table.entries[i].symbol, label) == 0)
        {
            return symbol_table.entries[i].location;
        }
    }

    return 0;
}

/**************************************************************************************************\
*   FUNCTION NAME:  write_objfile_header
****************************************************************************************************
*
*   Description:    Writes the header of the object file
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  None
*
*   Exceptions:     None
*
\**************************************************************************************************/
void write_objfile_header()
{
    objfile_hdr hdr;
    memset(&hdr, 0, sizeof(objfile_hdr));
    hdr.size_cs = num_instructions * MAX_INSTR_LENGTH;
    fseek(obj_file, 0, SEEK_SET);
    fwrite(&hdr, sizeof(objfile_hdr), 1, obj_file);
}

/**************************************************************************************************\
*   FUNCTION NAME:  parse_code
****************************************************************************************************
*
*   Description:    Parses the assembler input file
*
*   Arguments:      None
*
*   Return Value:   None
*
*   Special Notes:  * Writes the output object file
*
*   Exceptions:     None
*
\**************************************************************************************************/
void parse_code()
{
    // First pass
    get_segment_locations();
    location = OBJFILE_HEADER_LENGTH;

    parse_data_segment();
    parse_code_segment_pass1();

    // One and a half pass :)
    parse_code_segment_pass2();

    // Write header
    write_objfile_header();
}
