#define AX registers[0]
#define BX registers[1]
#define CX registers[2]
#define DX registers[3]
#define EX registers[4]
#define FX registers[5]
#define GX registers[6]
#define HX registers[7]
#define TIME_CYCLE 1
#define TIME_DONE 0
#define MAX_PROCESSES 20
#define MAX_SIMUL_PROCESSES 10
#define MEMORY_SIZE_IN_MB 1


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "emudefs.h"
#include "emulator.h"

// Emulator 
byte* DS, *CS;
short registers[NUM_REGISTERS], flags[NUM_FLAGS], IP;
int PID, processes_created = 0;
proc_table process_table;

void add_process_to_table(process* new_process);
void load_process(process* proc);
void save_process(process* proc);
void schedule();

void init_emulator()
{
    // Clear registers & flags
    memset(registers, 0, sizeof(registers) / sizeof(short));
    memset(flags, 0, sizeof(flags) / sizeof(short));

    // Initialize DS & CS pointers
    DS = NULL;
    CS = NULL;

    // Set IP to 0
    IP = 0;

    // Clear process table
    memset(&process_table, 0, sizeof(proc_table));
}

void init_process(char* filename)
{
    FILE* objfile = fopen(filename, "r");

    if (objfile != NULL)
    {
        // Get object file header
        objfile_hdr header;
        fread(&header, sizeof(header), 1, objfile);

        // Read program header (initalize IP, DS, CS)
        byte* data_segment = (byte*)malloc(header.size_ds);
        byte* code_segment = (byte*)malloc(header.size_cs);

        // Load data & code segments
        fread(data_segment, header.size_ds, 1, objfile);
        fread(code_segment, header.size_cs, 1, objfile);

        // Create new process space
        process* new_process = (process*) malloc(sizeof(process));

        // Clear new process memory
        memset(new_process, 0, sizeof(process));

        // Initalize process data
        new_process->state = WAITING;
        new_process->DS    = data_segment;
        new_process->CS    = code_segment;
        new_process->IP    = 0;
        new_process->PID   = ++processes_created;

        // Add entry to process table
        add_process_to_table(new_process);

        // Close file
        fclose(objfile);
    }
    else
    {
        // Print error message
    }
}

void add_process_to_table(process* new_process)
{
    if (process_table.num_curr_processes == 0)
    {
        process_table.first_process = new_process;
        process_table.curr_process = new_process;
        process_table.last_process = new_process;
    }
    else
    {
        // Find process to chain to
        process* curr_process = process_table.first_process;
        while(1)
        {
            if (curr_process->next == process_table.first_process)
            {
                curr_process->next = new_process;
                break;
            }
            else
            {
                curr_process = curr_process->next;
            }
        }
    }

    // Add to list of processes
    new_process->next  = process_table.first_process;
    process_table.num_curr_processes++;
}


int main(int argc, char *argv[])
{
    if (argc < 2)
    {
        fprintf(stderr, "Must specify a .obj file\n");
        fprintf(stderr, "Usage %s filename1.obj filename2.obj ...\n", argv[0]);
        return 1; // Exit program
    }
    else
    {
        init_emulator();


        int i;
        for (i=1; i < argc ;i++ )
        {
            init_process(argv[i]);
        }

        schedule();
        while(process_table.num_curr_processes != 0)
        {
            int remaining_cycles = 1;

            while(remaining_cycles != 0)
            {
                byte* instr = &CS[IP];

                if (instr[0] == HLT)
                {
                    process_table.curr_process->state = STOPPED;
                    break;
                }

                switch (instr[0])
                {
                    case ADD:
                    {
                        ADD(&registers[instr[1]], &registers[instr[2]]);
                        break;
                    }
                    case ADDI:
                    {
                        ADD(&registers[instr[1]], (short*)&instr[2]);
                        break;
                    }
                    case AND:
                    {
                        AND(&registers[instr[1]], &registers[instr[2]]);
                        break;
                    }
                    case DEC:
                    {
                        DEC(&registers[instr[1]]);
                        break;
                    }
                    case INC:
                    {
                        INC(&registers[instr[1]]);
                        break;
                    }
                    case JMP:
                    {
                        IP = instr[1];
                        break;
                    }
                    /*case MOV:
                    {
                        MOV(
                    }*/
                    case NOT:
                    {
                        NOT(&registers[instr[1]]);
                        break;
                    }
                    case OR:
                    {
                        OR(&registers[instr[1]], &registers[instr[2]]);
                        break;
                    }
                    case SUB:
                    {
                        SUB(&registers[instr[1]], &registers[instr[2]]);
                        break;
                    }
                    case SUBI:
                    {
                        SUB(&registers[instr[1]], (short*)&instr[2]);
                        break;
                    }
                    case XOR:
                    {
                        XOR(&registers[instr[1]], &registers[instr[2]]);
                        break;
                    }
                }

                printf("PID:=%d\n", PID);
                printf("R0:=%d; R1:=%d; R2:=%d; R3:=%d\n", AX, BX, CX, DX);
                printf("R4:=%d; R5:=%d; R6:=%d; R7:=%d\n", EX, FX, GX, HX);

                // Increment instruction pointer
                IP += MAX_INSTR_LENGTH;
                remaining_cycles--;
            }
            schedule();
        }
    }
}


void schedule()
{
    if (process_table.curr_process != NULL)
    {
        switch(process_table.curr_process->state)
        {
            case WAITING:
            {
                // Load
                load_process(process_table.curr_process);
                process_table.curr_process = process_table.curr_process;
                process_table.curr_process->state = RUNNING;
                break;
            }
            case RUNNING:
            {
                if (process_table.num_curr_processes > 1)
                {
                    // Perform context switch
                    save_process(process_table.curr_process);
                    load_process(process_table.curr_process->next);

                    // Save this process as last one to be run
                    process_table.last_process = process_table.curr_process;

                    // Make next process the current one
                    process_table.curr_process = process_table.curr_process->next;
                    process_table.curr_process->state = RUNNING;
                }
                break;
            }
            case STOPPED:
            {
                process* old_process = process_table.curr_process;

                // Perform clean up
                process_table.num_curr_processes--;

                // Connect previous and next process
                process_table.last_process->next = process_table.curr_process->next;

                // Return memory to system
                free(process_table.curr_process->DS);
                free(process_table.curr_process->CS);
                free(process_table.curr_process);

                // Make next process the current one
                process_table.curr_process = process_table.last_process->next;

                if (process_table.num_curr_processes != 0)
                {
                    schedule();
                }
                else
                {
                    process_table.curr_process == NULL;
                }

                break;
            }
        }
    }
}

//Time sharing STATE load
void load_process(process* proc)
{
    PID  = proc->PID;
    DS   = proc->DS;
    CS   = proc->CS;
    IP   = proc->IP;

    // Copy from PCB table into working memory
    memcpy(registers, proc->registers, sizeof(short) * NUM_REGISTERS);
    memcpy(flags, proc->flags, sizeof(short) * NUM_FLAGS);
}

//Time sharing STATE save
void save_process(process* proc)
{
    // Save instruction pointer
    proc->IP = IP;

    // Copy from working memory back into PCB table
    memcpy(proc->registers, registers, sizeof(short) * NUM_REGISTERS);
    memcpy(proc->flags, flags, sizeof(short) * NUM_FLAGS);
}

void DEC(short* dest)
{
    *dest--;
}

void INC(short* dest)
{
    *dest++;
}

void NOT(short* dest)
{
    *dest = ~*dest;
}

void XOR(short* dest, short* src)
{
    *dest = *dest ^ *src;
}

void OR(short* dest, short* src)
{
    *dest = *dest | *src;
}

void AND(short* dest, short* src)
{
    *dest = *dest & *src;
}

void ADD(short* dest, short* src)
{
    *dest = *dest + *src;
}

void SUB(short* dest, short* src)
{
    *dest = *dest - *src;
}